
#ifndef _BASE_DAG_HXX
#define _BASE_DAG_HXX 1

#include <memory.h>

typedef void  (*PF_TSSI)(void*pObj, void*pBuf, int iSet, int iInd);
typedef void *(*PF_TSIT)(void*pObj, void*pBuf);


/*
 *	tsort()
 *	topological sort of DAG.
 *
 *	pBase	pointer to a list of <c> nodes each of size<s>
 *		The DAG may have unconnected subgraphs.
 *
 *	s	sizeof(node)
 *	pBufA,
 *	pBufB	integer buffers of at least <c> elements
 *	c	node count of graph
 *	pFs	function called to start iteration over successor set of pObj
 *		pObj	elem of pBase
 *		pBuf	pIter, if nonzero start iteration
 *		iSet	if (iSet >= 0) then iSet is the order number to be assigned
 *			to the node
 *		iInd	if (iInd >= 0) then iInd is the number of 
 *			incoming arcs of the node
 *
 *	pFi	function called to iterate successor set
 *		pObj	elem of pBase
 *		pBuf	pIter	
 *		<return> pointer to the next successor node or 0.
 *			Note: Any pointer returned must address an element of pBase.
 *	
 *	pIter   iteration buffer
 *
 *	tsort() returns nonzero if the graph has cycles, otherwise
 *	any node has been applied a sequence number s. For any two
 *	nodes connected by an directed arc(a,b), s(a) < s(b).
 *
 *
 *	*** NOTES ***
 *
 *	If the DAG has multiple unconnected subgraphs, 
 *	the sequence numbers are freely distributed among them.
 *
 *	tsort() does not sort the elements of the node array.
 *	In order to sort the node array itself, the user must qsort() the node
 *	array by the comparision of the sequence numbers.
 */

int tsort(void *pBase, size_t s, int *pBuf1, int *pBuf2, int c, PF_TSSI pFs, PF_TSIT pFi, void *pIter);

#endif
