#ifndef _UMI_UMI_HXX
#define _UMI_UMI_HXX

/**
 * \file 
 * Universal Message Interface - Generic C++ part
 */
#include "umi.h"
#include <limits.h>
#if (defined(MMSYS_linux) || defined(MMSYS_lkern))
#include <sys/time.h>
#else
  struct timespec {
    unsigned tv_sec;
    unsigned tv_nsec;
  };
#endif

#ifndef _MM_NO_IOSTREAM
 #include <iostream>
#endif

#ifndef EOF 
  #define EOF (-1)
#endif



class xostream;
class xistream;

namespace INdepUmi100 {

  /**
   * base class for the control part of messages
   */
  class UmiControl:public UmiControlStruct {
   public:
    bool isType(unsigned e) { return _eType == e; }
    int eType() const { return _eType; }
    size_t size() const { return _size; }

    //size_t sizeData() const { return _pchLim-_pchBase; }
    friend xostream &xmarshal(xostream&, const UmiControl &) ;
    friend xistream &xunmarsh(xistream &, UmiControl &);

    //void iLength(int) {}
  #ifndef _MM_NO_IOSTREAM
    friend std::ostream &operator<<(std::ostream &, const UmiControl &);
  #endif
    friend xistream &operator>>(xistream &, UmiControl &);
    friend bool operator==(const UmiControl &, const UmiControl&);	/**< binary comparison */
    UmiMemStruct um() const;

    //UmiControl() { _eType=0, _size=0, _ud._pchBase=_ud._pchLimDat=_ud._pchLimBuf=0; }
    //UmiControl(int eType, size_t s)  { _eType=eType, _size=s, _ud._pchBase=_ud._pchLimDat=_ud._pchLimBuf=0; }
    UmiControl() { _eType=0, _size=0 ; }
    UmiControl(int eType, size_t s)  { _eType=eType, _size=s ; }
    ~UmiControl() { }
  };

  /**
   * user level data descriptor, used to pass buffer information
   * the UM interface
   */
  class UmiData:public UmiMemStruct {
   public:
    //size_t sizeMax() const { return SZ_DATA_MAX ; }
    char *pchBase() const { return _pchBase; }
    char *pchLimBuf() const { return _pchLimBuf; }
    char *pchLimDat() const { return _pchLimDat; }
    void setSize(const UmiMemStruct &um) { _pchLimDat = _pchBase + (um._pchLimDat-um._pchBase); }
    size_t cData() const { return _pchLimDat-_pchBase; }
    size_t cMax() const { return _pchLimBuf-_pchBase; }


    UmiData() {
      _pchLimDat=_pchLimBuf=_pchBase=0;
    }
    UmiData(char *pch, size_t s) {
      _pchBase=pch;
      _pchLimDat=_pchLimBuf=pch+s;
    }
    ~UmiData() { }
  };

  /**
   * the control part receive buffer 
   */
  class UmiControlBuf:public UmiControl {
    char _mpch[SZ_CTRL_MAX];
  public:
    UmiControlBuf& operator=(const UmiControl &); /**< copy in any UmiControl */
    UmiControlBuf();
    ~UmiControlBuf();
  };



  /**
   * the data part receive buffer.
   * defines a buffer of SZ_DATA_MAX chars
   */
  class UmiDataBuf:public UmiData {
    char _mpch[ __SZ_DATA_MAX ];
   public:
    UmiDataBuf():UmiData(_mpch, __SZ_DATA_MAX) {}
    ~UmiDataBuf() {}
  };

  /**
   * Posix 1b timespec c++ wrapper class
   */
  class Time:public timespec {
   public:
    friend Time operator+( const Time &tiL, const Time &tiR );
    friend bool operator>=( const Time &tiL, const Time &tiR );
    friend bool operator==( const Time &tiL, const Time &tiR );
    friend xistream &xunmarsh( xistream &, Time & );
    friend xostream &xmarshal( xostream &, const Time &) ;
    #ifndef _MM_NO_IOSTREAM
      friend std::ostream &operator<<(std::ostream &, const Time &);
    #endif


    Time() { tv_sec=0; tv_nsec=0; }   /* default time value is 0.0 seconds */
    Time( const Time &ti ) { tv_sec=ti.tv_sec; tv_nsec=ti.tv_nsec; }
    Time( const struct timespec &ti ) { tv_sec=ti.tv_sec; tv_nsec=ti.tv_nsec; }
    Time( unsigned sec, unsigned nsec=0 ) { tv_sec=sec; tv_nsec=nsec; }
    ~Time() {}
  };

  /*
   * special value to represent "infinity",
   * may only be used for the == operator
   */
  inline Time tiInf() {
    return Time(UINT_MAX, UINT_MAX);
  };


  /*
   * UmiGetBuf, UmiPutBuf -- universal put/get buffers
   */
  class UmiGetBuf {
    const char * const _pchBase;	/* first byte of allocation	 */
    const char * const _pchLimBuf;	/* last byte +1 of buffer	 */

    const char *_pchLimDat;		/* last byte +1 of valid data	 */
    const char *_pch;			/* next byte to get		 */

    int zapeof(const char *pch) { return *reinterpret_cast<const unsigned char *>(pch); }

   public:
    /* get */
    int cDone() { return _pch-_pchBase; }	/* chars done	   */
    int sgetc() { return (_pch >= _pchLimDat) ? EOF : zapeof(_pch); }
    int sbumpc() { return (_pch >= _pchLimDat) ? EOF : zapeof(_pch++); }
    int sbumpc_checked() { return zapeof(_pch++); }  /* optimized version */

    const char *pchBase() const { return _pchBase; }
    int cMax() const { return _pchLimBuf-_pchBase; }		/* chars max */
    void setCount(size_t c) { _pchLimDat=(_pch=_pchBase)+c; }

    UmiGetBuf(char *pchBase, size_t s): _pchBase(pchBase), _pchLimBuf(pchBase+s), _pchLimDat(pchBase), _pch(pchBase) { }
    ~UmiGetBuf() { }

  };

  class UmiPutBuf  {
    char * const _pchBase;	/* first byte of allocation	 */
    char * const _pchLimBuf;	/* last byte +1 of buffer	 */

    char *_pchLimDat;		/* last byte +1 of valid data	 */
    char *_pch;			/* next byte to put		 */
   public:
    /* put */
    bool canput() { return _pch < _pchLimDat; }
    bool canputnext() { return (_pch+1) < _pchLimDat; }
    void sputc( int ch ) { *(_pch++)=(unsigned char)ch; }
    size_t cDone() const { return _pch-_pchBase; }		/* chars done	   */
    char *pchBase() const { return _pchBase; }
    int cRem() { return _pchLimDat-_pch; }		/* chars remaining */

    /*
     * set the limit of bytes the driver accepts before
     * returning !canput(). This limit is independent of
     * the buffer size.
     */
    void setLimit(size_t s) { _pch=_pchBase, _pchLimDat=_pchBase+s; }

    UmiPutBuf(char *pchBase, size_t s): _pchBase(pchBase), _pchLimBuf(pchBase+s), _pchLimDat(pchBase), _pch(pchBase) { }
    ~UmiPutBuf() { }
  };

  /**
   * protocol information 
   *
   * This class defines the protocol spoken between the UMI user and
   * the driver. It also specifies the wire format of the protocol messages and
   * thus can be used to configure universal proxies.
   */
  class UmiProto /*:public UmiProtoStruct*/ {
    const char *_szProto;	
    const char *_szName;
   public:
    /**
     * unmarshal (i.e.\ convert from the wire data to the internal format and byte order, respectively) message.
     * This pure virtual must be defined by the derived class.
     *
     * \param xi xistream to fetch the wire data from
     * \param uc header of the control message
     * \param puc pointer to the buffer for the converted message
     * \return the xi argument
     */
    virtual xistream& unmarsh(xistream &xi, const UmiControl &uc, UmiControl *puc) const;

    /**
     * marshal (i.e.\ convert from the internal format and byte order, respectively, to the wire data) message.
     * This pure virtual must be defined by the derived class.
     *
     * \param xo xostream to write the wire data to
     * \param puc pointer to the message to be marshalled
     * \return the xo argument
     */
    virtual xostream& marshal(xostream &xo, const UmiControl *puc) const ;

    const char *szProto() const { return _szProto; }	/**< protocol descripion */
    const char *szName() const { return _szName; }	/**< protocol name	 */

    UmiProto(const char *szProto, const char *szName);
    virtual ~UmiProto();

    /**
     * check for protocol identity
     */
    static bool match(const char *szL, const char *szR);
    friend bool operator==(const UmiProto &upL, const UmiProto &upR);
  };
}

#endif



