#ifndef _ADIAD_H_
#define _ADIAD_H_

#include "fixme.h"

 /**
  * ADI adaptor descriptors
  */

  /**
   * adaptor base class.
   * Same for all adaptors.
   */
  typedef struct {
    const char *_szName;	/**< Adaptor name		*/
    const char *_szOrg;		/**< organization		*/
    const char *_szArch;	/**< Adaptor architecture 	*/
    const char *_szLic;		/**< License string or 0	*/
  } AdiAdapt; 


  /* 
   * flags for PCI region descriptors 
   * each region listed is mapped
   */
  #define RG_AMCC5920_CONF 001	/* AMCC local config region	*/
  #define RG_IOBASE 002		/* GPIB register region		*/
  #define RG_IOMEM 004		/* GPIB memory region		*/
  #define RG_ISMEM(x) (((x) & RG_IOMEM)!=0)

  #define MAX_PCI_REGION 2
  #define MAX_REGION MAX_PCI_REGION 

  /**
   * PCI bus specific region descriptor
   */
  typedef struct {
    int _iBar;			/**< PCI bar offset, -1 = NIL	*/
    int _f;			/**< what to do with this region	*/
    int _l;			/**< region length in bytes	*/
  } RegionDesc ;

  /**
   * PCI bus specific adaptor descriptor.
   */
  typedef struct {
    AdiAdapt _ad;		/**< base class	*/
    void *_pArg;		/**< argument for the UMI constructor */

    struct {
      uint32_t _vid;		/**< vendor id		*/
      uint32_t _did;		/**< device id		*/
    } _id, _idSS;

    RegionDesc _mprd[MAX_PCI_REGION];	/**< region descriptors */
  } AdiAdaptPci; 

  typedef struct {
    AdiAdapt _ad;
    uint16_t _vid;
    uint16_t _pid;
    const char *_szMcuType;		/**< type if the MCU			*/
    const void *_mppfw[MAX_DRIV];	/**< per driver firmware download data	*/
  } AdiAdaptUsb; 

  struct LumiModStruct;

  typedef struct {
    AdiAdapt _ad;
    void *_pArg;		/**< argument for the UMI constructor */
    const int *_piSig;	/* signature[_isaLen]		*/
    int _isaLen;
    int _f;
  } AdiAdaptIsa; 

#endif
