
#include <linux/types.h>
#include "fixme.h"
#include "umic/umi.h"
#include "card.h"
#include "mod.h"
#include "assert.h"
#include <linux/sched.h>
#include <linux/version.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>

 /*
  * bus.c
  *
  * driver for PCI and ISA cards 
  */

  /*
   * card status including interpreter 
   * virtual machine
   */
  struct Region {
    unsigned long _lAddr;
    unsigned long _lLen;
    int _f;
  };

  struct CardBusS;
  typedef struct CardBusS CardBus;

  struct ChanBusS {
    struct UmiDriver *_pud;
    struct UmiChannel *_puc;
    CardBus *_pcb;

    /* wait queues */
    wait_queue_head_t _wqDown;	/* waiting to put message	*/
    wait_queue_head_t _wqUp;	/* waiting to get message	*/
    const char *_szStat;
  };
  typedef struct ChanBusS ChanBus;
  #define Chan_stat(pcn, sz) (pcn)->_szStat=(sz)



  struct CardBusS {
    Card _ca;				/* header			*/

    struct UmiMemInfo _mpmi[MAX_UMI_MEM];  /* memory for the UMI card driver */
    int _cmi;
    struct UmiCard *_puc;

    ChanBus _mpcn[MAX_CACN];
//    int _ccn;

    spinlock_t _slCard;			/* thread lock mutex		*/
    int _fCheck;
    int _fOpen;				/* REMIND: must be ATOMIC open channel flag		*/

    /* per channel wait queues */

    /* timer */
    struct timer_list _timer;		/* timer tick			*/
    atomic_t _fTimerStop;

    /*
     * Adaptor information
     */
    union {
      UmiAdaptIsa *_isa;
      UmiAdaptPci *_pci;
    } _pad;
    int _iIoRegion;			/* region for controller IO	*/
    int _iMemRegion;			/* region for memory access 	*/
    int _irq;
    int _fIrq;				/* shared/non shared		*/
    struct Region _mprg[MAX_PCI_REGION];

    /* ISA/PCI hardware description */
  } ;

  #define pvm(p) (((p)->_pim))
  #define pcb_pdm(p) (pca_pdm(&(p)->_ca))
  #define pcb_pad(p) (&((p)->_pad._pci->_ad))
  #define pcb_pcn(p, i) ((p)->_mpcn+(i))


  void Chan_debug(LumiMod *plm, ChanBus *pcn, char *szMsg, int iStat) {
    if (*plm->_pfDebug) printk(KERN_DEBUG PFX " %s:%s:%d %s %x\n", plm->_szDid, pdm_getid(plm, &pcn->_pcb->_ca),  UmiDriver_szName(pcn->_pud), UmiChannel_num(pcn->_puc), szMsg, iStat);
  }

  typedef unsigned long irq_flag_t;

  /* per card timer */
  void bustimer(unsigned long p) {
    CardBus *pcb=(CardBus*)p;
    irq_flag_t f;
    //char mz[80];

    spin_lock_irqsave(&pcb->_slCard, f);	
    //debugmsg(pif, "timer", GpibInterface_print(pif->_pgi, mz, sizeof(mz)));
    if (1) {
      struct timespec ts;
      unsigned long j=jiffies;		/* REMIND lock		*/
      ts.tv_sec=j/HZ;
      ts.tv_nsec=(j%HZ)*(1000000000/HZ);

      UmiCard_timer(pcb->_puc, 333, &ts);
    } else {
      printk(KERN_DEBUG PFX "timer call  SKIPPED\n", pcb_pdm(pcb)->_szDid);
    }
    spin_unlock_irqrestore(&pcb->_slCard, f);	
     

    /* reschedule timer */
    pcb->_timer.expires=jiffies + (HZ/pcb_pdm(pcb)->_pum->_tps) ;
    if (!atomic_read(&pcb->_fTimerStop)) add_timer(&pcb->_timer);
  }

  /*
   * per card interrupt
   */
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  void busintr(int irq ,void * dev_id, struct pt_regs *regs) {
 #else
  irqreturn_t busintr(int irq ,void * dev_id, struct pt_regs *regs) {
 #endif
    CardBus *pcb=(CardBus*)dev_id;


    /* run the interface interrupt procedure */
    irq_flag_t f;
    spin_lock_irqsave(&pcb->_slCard, f);	

    /* _all_ other (timer) local interrupts are off now */

    if (pcb->_fCheck) panic("lumi interrupt recursion\n");
    pcb->_fCheck=1;
    
    assert(pcb->_puc);
    UmiCard_interrupt(pcb->_puc, 666);
    pcb->_fCheck=0;
    spin_unlock_irqrestore(&pcb->_slCard, f);	
     

    #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
      return IRQ_HANDLED;
    #endif
  } 



  #define SZ_CTRL_HDR (sizeof(struct UmiDataStruct))

  char* busPrint( Card *pca, char* pchBuf, int c) {
    CardBus *pcb=(CardBus*)pca;
    char *pch=pchBuf;
    //char *pchLim=pchBuf+c-80;
    char mpch[80];

    UmiCard_print(pcb->_puc, mpch, 80);
    pch += sprintf(pch, "card=%s (%s) ", pcb_pad(pcb)->_szName, mpch);
    pch += sprintf(pch, "io=%lx irq=%d", pcb->_mprg[pcb->_iIoRegion]._lAddr , pcb->_irq);
    if (pcb->_iMemRegion != -1) pch += sprintf(pch, " mem=%lx", pcb->_mprg[pcb->_iMemRegion]._lAddr);
    return pch;
  }

  /*
   * put data down
   */
  int bufPutDown( struct UmiMemStruct *pudU, char *pch, size_t sMax, size_t *psCur) {
    char *pchBase, *pchLim;
    size_t sCur;
    __get_user(pchBase, &pudU->_pchBase);
    __get_user(pchLim, &pudU->_pchLimDat);
    sCur=pchLim-pchBase;
    if (sCur > sMax) {
      printk(KERN_ERR "bufPutDown %p %p %zu > %zu\n", pchBase, pchLim,  sCur, sMax);
      return -EFAULT;
    }
    if (copy_from_user(pch, pchBase, sCur)) {
      printk(KERN_ERR "bufPutDown copy %p %p %zu\n",  pch, pchBase, sCur);
      return -EFAULT;
    }
    *psCur=sCur;

    return 0;
  }

  /*
   * get data up
   * copy s bytes from pchC to the user buffer described by pudU
   */
  int bufGetUp( struct UmiMemStruct *pudU, char *pchC, size_t s) {
    char *pchBaseU;
    char *pchLimU;
    if (get_user(pchBaseU, &pudU->_pchBase)) return -100;
    if (get_user(pchLimU, &pudU->_pchLimBuf)) return -101;
    if ((pchLimU-pchBaseU) < s) {
      printk(KERN_ERR "get_user failed, %td %zu\n",  pchLimU-pchBaseU, s);
      return -102;
    }
    if (copy_to_user(pchBaseU, pchC, s)) return -103;
    __put_user(pchBaseU+s, &pudU->_pchLimDat);
    return 0;
  }

  int Chan_canput(ChanBus *pcn) { 
    int r;
    irq_flag_t f;
    spin_lock_irqsave(&pcn->_pcb->_slCard, f);
    r=UmiChannel_canput(pcn->_puc); 
    spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    return r;
  }

  int Chan_canget(ChanBus *pcn) { 
    int r;
    irq_flag_t f;
    spin_lock_irqsave(&pcn->_pcb->_slCard, f);
    r=UmiChannel_canget(pcn->_puc); 
    spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    return r;
  }

  int busPutDown(Card *pca, Chan *pcc, unsigned long arg) {
    int iChan=pcc->_iChanOfCard;
    CardBus *pcb=(CardBus*)pca;
    ChanBus *pcn=pcb_pcn(pcb, iChan);
    struct UmiMsgDownBuf *pdb;
    UmiIoc *pubU=(void*)arg;
    int iErr;
    size_t s;
    assert(iChan >= 0 && iChan < MAX_CACN);

    //printk(KERN_DEBUG "busPutDown(\n");
    /* wait here until ready */
    iErr = wait_event_interruptible(pcn->_wqDown, (Chan_canput(pcn)));
    if (iErr) return iErr;

    {
      /* allocate a downstream buffer */
      irq_flag_t f;
      spin_lock_irqsave(&pcn->_pcb->_slCard, f);
      pdb=UmiChannel_allocDownBuf(pcn->_puc);
      spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    }

    Chan_stat(pcn, "ioctl-put-alloc");

    /* verify and copy in the control part */
    if (bufPutDown(&pubU->_cd, (char*)UmiMsgDownBuf_pControl(pdb), SZ_CTRL_MAX, &s)) return -EFAULT;
    if (bufPutDown(&pubU->_dd, UmiMsgDownBuf_pData(pdb), UmiMsgDownBuf_sizeDataMax(pdb), &s)) return -EFAULT;
    UmiMsgDownBuf_setDataLim(pdb, s);

    Chan_stat(pcn, "ioctl-put-copied");

    /* 
     * send the message downstream 
     */
    {
      irq_flag_t f;

      spin_lock_irqsave(&pcn->_pcb->_slCard, f);

      Chan_debug(pcb_pdm(pcb), pcn, "put", 0);
      UmiChannel_putDownBuf(pcn->_puc, 0, pdb);

      spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    }
    Chan_stat(pcn, "ioctl-put-done");
    //printk(KERN_DEBUG "busPutDown)\n");

    return 0;
  }

  /**
   * get upstream message from the UMI core
   */
  int busGetUp(Card *pca, Chan *pcc, unsigned long arg) {
    int iChan=pcc->_iChanOfCard;
    CardBus *pcb=(CardBus*)pca;
    ChanBus *pcn=pcb_pcn(pcb, iChan);
    int e;
    int iErr;
    UmiIoc *pubU=(void*)arg;
    struct UmiMsgUpBuf *pub;

    assert(iChan >= 0 && iChan < MAX_CACN);

    Chan_stat(pcn, "ioctl-get");
    /* wait here until ready */
    iErr = wait_event_interruptible(pcn->_wqDown, Chan_canget(pcn));
    if (iErr) {
      Chan_stat(pcn, "ioctl-get-intr");
      return iErr;
    }
    Chan_stat(pcn, "ioctl-get-ok");

    /* get the message */
    {
      irq_flag_t f;
      spin_lock_irqsave(&pcn->_pcb->_slCard, f);

      Chan_debug(pcb_pdm(pcb), pcn, "get", 0);
      pub=UmiChannel_getUpBuf(pcn->_puc);

      spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    }
    Chan_stat(pcn, "ioctl-get-buf");

    /*
     * save the pointers to the user data buffer
     * REMIND: How to return (set) the user data limit ?
     */
    assert(pub);
    //printk(KERN_DEBUG PFX "getup control %zu\n", pcb_pdm(pcb)->_szDid, UmiMsgUpBuf_sizeControl(pub));
    if ((e=bufGetUp(&pubU->_cu, (char*) UmiMsgUpBuf_pControl(pub), UmiMsgUpBuf_sizeControl(pub)))) return e;
    //printk(KERN_DEBUG PFX "getup data %zu\n", pcb_pdm(pcb)->_szDid, UmiMsgUpBuf_sizeData(pub));
    if ((e=bufGetUp(&pubU->_du, UmiMsgUpBuf_pData(pub), UmiMsgUpBuf_sizeData(pub)))) return e;

    Chan_stat(pcn, "ioctl-get-copied");

    /* release the buffer */
    {
      irq_flag_t f;
      spin_lock_irqsave(&pcn->_pcb->_slCard, f);

      Chan_debug(pcb_pdm(pcb), pcn, "release", 0);
      UmiChannel_releaseUpBuf(pcn->_puc, 0, pub);

      spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    }

    Chan_stat(pcn, "ioctl-get-done");
    return 0;
  }

  int busOpen(Card *pca, Chan *pcc) {
    int iChan=pcc->_iChanOfCard;
    CardBus *pcb=(CardBus*)pca;
    register ChanBus *pcn=pcb_pcn(pcb, iChan);
    int e;
    assert(iChan >= 0 && iChan < MAX_CACN);
    //printk(KERN_DEBUG " busOpen %p %p %p\n", pca, pcn, pcn->_puc);
    {
      irq_flag_t f;
      spin_lock_irqsave(&pcn->_pcb->_slCard, f);
      e=UmiChannel_open(pcn->_puc);
      spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    }
    return e ? -EIO : 0;
  }


  int busClose(Card *pca, Chan *pcc) {
    int iChan=pcc->_iChanOfCard;
    CardBus *pcb=(CardBus*)pca;
    register ChanBus *pcn=pcb_pcn(pcb, iChan);
    assert(iChan >= 0 && iChan < MAX_CACN);
    {
      irq_flag_t f;
      spin_lock_irqsave(&pcn->_pcb->_slCard, f);
      UmiChannel_close(pcn->_puc);
      spin_unlock_irqrestore(&pcn->_pcb->_slCard, f);
    }
    return 0;
  }



  /*
   * kmalloc all memory required
   */
  int kmalloc_all( struct UmiMemInfo *pmi, int c) {
    int i;
    for (i=0; i<c; i++, pmi++) {
      pmi->_p=kmalloc(pmi->_s, GFP_KERNEL);
      if (!pmi->_p) return -ENOMEM;
    }
    return 0;
  }

  int kfree_all( struct UmiMemInfo *pmi, int c) {
    int i;
    for (i=0; i<c; i++, pmi++) if (pmi->_p) kfree(pmi->_p);
    return 0;
  }

  /*
   * Card construction/destruction
   */
  CardBus *busCardDelete(LumiMod *pdm, CardBus *pcb) {
    if (pcb) {
      kfree_all(pcb->_mpmi, pcb->_cmi);
      kfree (pcb);
    }
    return 0;
  }

  /*
   * signal procs called from UMI channel
   */
  void buscanget( int iCtx , void *p) {
    ChanBus *pcn=(ChanBus *)p;
    assert(pcn);
    wake_up_interruptible(&pcn->_wqUp);
  }

  void buscanput( int iCtx , void *p) {
    ChanBus *pcn=(ChanBus *)p;
    assert(pcn);
    wake_up_interruptible(&pcn->_wqDown);
  }

  /*
   * tiNext function
   */
  void tiNextTick( struct timespec *ptsNext ) {
    unsigned long j=jiffies+1;		/* REMIND lock	*/
    ptsNext->tv_sec=j/HZ;
    ptsNext->tv_nsec=(j%HZ)*(1000000000/HZ);
  }

  struct UmiEnvC ue = { tiNextTick, buscanget, buscanput } ; 

  /*
   * new card for module pdm->_pum<pArg>
   */
  CardBus *busCardNew(LumiMod *pdm) {
    int i;
    ChanBus *pcn;
    CardBus *pcb=(CardBus*)kmalloc(sizeof(CardBus), GFP_KERNEL);
    if (!pcb) return 0;

    /*
     * malloc UMI kernel memory
     */
    pcb->_puc=0;
    pcb->_cmi=UmiCard_getMem(pdm->_pum, pcb->_mpmi, MAX_UMI_MEM);
    if (pcb->_cmi < 0) {
      printk(KERN_ERR PFX "can't get memory information, ret=%d, buffer max=%d\n", pdm->_szDid, pcb->_cmi, MAX_UMI_MEM);
      return busCardDelete(pdm, pcb);
    }
    i=kmalloc_all(pcb->_mpmi, pcb->_cmi);
    if (i) {
      printk(KERN_ERR PFX "can't kmalloc() %d blocks\n", pdm->_szDid, pcb->_cmi);
      return busCardDelete(pdm, pcb);
    }
    else {
      printk(KERN_DEBUG PFX "done kmalloc() %d blocks\n", pdm->_szDid, pcb->_cmi);
    }


    pca_init(&pcb->_ca, pdm );
    pcb->_ca._pfPutDown=busPutDown;
    pcb->_ca._pfGetUp=busGetUp;
    pcb->_ca._pfOpen=busOpen;
    pcb->_ca._pfClose=busClose;
    pcb->_ca._pfPrint=busPrint;

    spin_lock_init(&(pcb)->_slCard);
    pcb->_fCheck=0;
    pcb->_fOpen=0;

    pcn=pcb->_mpcn;
    for (i=0; i<MAX_CACN; i++, pcn++) {
      pcn->_pud=0;
      pcn->_puc=0;
      pcn->_pcb=pcb;
      init_waitqueue_head(&pcn->_wqUp);
      init_waitqueue_head(&pcn->_wqDown);
      pcn->_szStat="";
    }
//    pcb->_ccn=0;

    atomic_set(&pcb->_fTimerStop, 0);
    init_timer(&pcb->_timer);
    pcb->_timer.function=bustimer;
    pcb->_timer.data=(unsigned long)pcb;
    pcb->_timer.expires=jiffies + (HZ/pcb_pdm(pcb)->_pum->_tps) ;


    pcb->_pad._isa=0;
    pcb->_iIoRegion=-1;
    pcb->_iMemRegion=-1;
    pcb->_irq=0;
    pcb->_fIrq=0;
    memset(pcb->_mprg, 0, sizeof(pcb->_mprg));


    return pcb;
  }

  void busCardInitUmi(LumiMod *pdm, CardBus *pcb, void *pArg) {
    /*
     * create UMI card, driver, and channels
     */
    pcb->_puc=UmiCard_new(pdm->_pum, &ue, pcb->_mpmi, pcb->_cmi, pArg, pcb->_mprg[pcb->_iIoRegion]._lAddr);
    printk(KERN_DEBUG PFX "umi: new card %p <%p>\n", pdm->_szDid, pcb->_puc, pArg);

    {
      //ChanBus *pcn=pcb->_mpcn;
      int iDriver=0;

      struct UmiDriver *pud;
      struct UmiIter it;
      UmiCard_startIterDrivers(pcb->_puc, &it);
      while ((pud=UmiCard_iterateDrivers(pcb->_puc, &it))) {
	struct UmiIter it;
	struct UmiChannel *puc;

	int iChanOfDriver=0;

	UmiDriver_startIterChannels(pud, &it);
	while ((puc=UmiDriver_iterateChannels(pud, &it))) {
	  int iChan=iDriver*MAX_DRCN-iChanOfDriver;

	  if (iChan < MAX_CACN) {
	    ChanBus *pcn=pcb->_mpcn+iChan;
	    /* init channel */
	    pcn->_pud=pud;
	    pcn->_puc=puc;
	    pcn->_pcb=pcb;
	    UmiChannel_setHandle(pcn->_puc, pcn);		/* for callback	*/

	    init_waitqueue_head(&pcn->_wqUp);
	    init_waitqueue_head(&pcn->_wqDown);

	    printk(KERN_DEBUG PFX "umi: new chan %td %s %d\n", pdm->_szDid, pcn-pcb->_mpcn, UmiDriver_szName(pcn->_pud), UmiChannel_num(pcn->_puc));
	  }
	  iChanOfDriver++;
	}
        iDriver++;
      }
    }
  }





  /*
   * ISA (init/fini)ialization 
   */

  static char * szIdentIsa( char *mpchBuf, int io ) {
    sprintf(mpchBuf, "isa:%x", io);
    return mpchBuf;
  }


  int enum_isa_devices(LumiMod *pdm) {
    int i;

		  
    /* Check for user specified ISA devices */
    for (i=0 ; pdm->_padIsa && (i < pdm->_cMaxIsa) && pdm->_piIoIsa[i] && pdm->_piIrqIsa[i]; i++) {
      char mpch[SZ_CARD_ID];
      int iSlot;
      CardBus *pcb;
      void *pArg;

      if (*pdm->_pfDebug) printk(KERN_INFO PFX "ISA device specified io=%04X,irq=%d\n", pdm->_szDid, pdm->_piIoIsa[i], pdm->_piIrqIsa[i] ); 

      if ((iSlot=pdm_getslot(pdm, szIdentIsa(mpch, pdm->_piIoIsa[i])))==-1) return -1;

      /*
       * allocate and initialize a card structure
       */
      pcb=busCardNew(pdm);
      if (!pcb) return -1;

      /* begin init ISA fields */
      {
	UmiAdaptIsa *pad;		/* current adaptor */
	struct Region *prg=pcb->_mprg+0;
	int io=pdm->_piIoIsa[i];
	int irq=pdm->_piIrqIsa[i];
	pad=pcb->_pad._isa=pdm->_padIsa;
	pArg=pad->_pArg;

	pcb->_iIoRegion=0;
	pcb->_iMemRegion=-1;

	printk(KERN_INFO PFX "try IO region %x(%d)\n", pdm->_szDid, io, pad->_rd._l);

	if (request_region(io, pad->_rd._l, pdm->_szDid)==0) {
	  printk(KERN_ERR PFX "IO region %x(%d) not available\n", pdm->_szDid, io, pad->_rd._l);
	  return (busCardDelete(pdm, pcb), -ENODEV);
	}

	prg->_lAddr=io;
	prg->_lLen=pcb->_pad._isa->_rd._l;
	prg->_f=IORESOURCE_IO;

	printk(KERN_INFO PFX "adaptor type=%s manuf=%s arch=%s\n",pdm->_szDid, pad->_ad._szName, pad->_ad._szOrg, pad->_ad._szArch);

	/* request irq */
	pcb->_irq=irq;
	pcb->_fIrq=0;
	printk(KERN_INFO PFX "IRQ %d\n", pdm->_szDid, pcb->_irq);
        if (request_irq(pcb->_irq, busintr, pcb->_fIrq, pdm->_szDid, pcb) ) {
	  printk(KERN_ERR PFX "IRQ %d not available\n", pdm->_szDid, pcb->_irq);
	  return (busCardDelete(pdm, pcb), -ENODEV);
        }
      }

      /* done init ISA fields */

      busCardInitUmi(pdm, pcb, pArg);

      /*
       * register the card
       */
      pdm_setcard(pdm, iSlot, &pcb->_ca);
      add_timer(&pcb->_timer);
    }
    return 0;
  }


  int free_isa_devices(LumiMod *pdm) {
    int i;
		  
    /* Check for user specified ISA devices */
    for (i=0 ;(i < pdm->_cMaxIsa) && pdm->_piIoIsa[i] && pdm->_piIrqIsa[i]; i++) {
      char mpch[SZ_CARD_ID];

      int iSlot;
      if ((iSlot=pdm_findslot(pdm, szIdentIsa(mpch, pdm->_piIoIsa[i])))!=-1) {
	CardBus*pcb;
	if ((pcb=(CardBus*)pdm_clrcard(pdm, iSlot))) {
	  printk(KERN_INFO PFX "release isa card %s@%d\n", pdm->_szDid, mpch, iSlot);
	  {
	    {
	      struct Region *prg=pcb->_mprg+0;
	      printk(KERN_INFO PFX " release region %lx %ld\n", pdm->_szDid, prg->_lAddr, prg->_lLen);

	      if (prg->_f & IORESOURCE_IO) release_region(prg->_lAddr,prg->_lLen);
	      else if (prg->_f & IORESOURCE_MEM) release_mem_region(prg->_lAddr,prg->_lLen);
	    }
          }
	  busCardDelete(pdm, pcb);
	}
      }
    }
    return 0;
  }

  /*
   * PCI probe/remove
   */

  static char * szIdentPci( char *mpchBuf, struct pci_dev *pdv ) {
    sprintf(mpchBuf, "pci:%s", pci_name(pdv));
    return mpchBuf;
  }

  void busReleasePci(CardBus *pcb) {
    int i;

    /* notify concurrent bustimer() not to add the timer again */
    atomic_set(&pcb->_fTimerStop,1);
    del_timer_sync(&pcb->_timer); 	
    if (pcb->_irq) free_irq(pcb->_irq, pcb);

    for(i=0; i<MAX_PCI_REGION; i++) {
      struct Region *prg=pcb->_mprg+i;

      if (prg->_f & IORESOURCE_IO) release_region(prg->_lAddr,prg->_lLen);
      else if (prg->_f & IORESOURCE_MEM) release_mem_region(prg->_lAddr,prg->_lLen);
    }
  }

  /*
   * probe() might be executed by indepent threads,
   * so the access to the card structure is locked.
   */
  int pdm_pci_probe( LumiMod *pdm, struct pci_dev *pdv, const struct pci_device_id *pdi) {
    char mpch[SZ_CARD_ID];
    int iSlot;
    void *pArg;
    CardBus *pcb;
    if (!pdi->driver_data) return -1;

    printk(KERN_INFO PFX "pci probe (%s)\n", pdm->_szDid, pci_name(pdv));

    if ((iSlot=pdm_getslot(pdm, szIdentPci(mpch, pdv)))==-1) return -1;

    /*
     * allocate and initialize a card structure
     */
    pcb=busCardNew(pdm);
    if (!pcb) return -1;

    {
      UmiAdaptPci *pad;		/* current adaptor */
      int i;
      if (pci_enable_device(pdv)) goto lFail;

      pad=pcb->_pad._pci=(UmiAdaptPci*)pdi->driver_data;
      pArg=pad->_pArg;


      /*
       * get regions 
       */
      for(i=0; i<MAX_PCI_REGION; i++) {
	unsigned long lAddr, lLen, lFlags;
	const RegionDesc *prd=pad->_mprd+i;
	struct Region *prg=pcb->_mprg+i;

	if (prd->_iBar == -1) break;	/* region not used */

	lAddr=pci_resource_start(pdv, prd->_iBar);
	lLen=pci_resource_len(pdv, prd->_iBar);
	lFlags=pci_resource_flags(pdv, prd->_iBar);

	if (!lAddr) {
	  printk(KERN_ERR PFX "PCI region #%d not available\n", pdm->_szDid, prd->_iBar);
	  goto lFail;
	}
	printk(KERN_INFO PFX "PCI Region %d addr=%lx len=%lx flags=%lx\n", pdm->_szDid, prd->_iBar, lAddr, lLen, lFlags);

	if (RG_ISMEM(prd->_f)) {
	  /* MEM space */
	  if (check_mem_region(lAddr,lLen)) {
	    printk(KERN_ERR PFX "memory region #%d not available\n", pdm->_szDid, prd->_iBar);
	    goto lFail;
	  }
	  request_mem_region(lAddr,lLen, pdm->_szDid);
	  prg->_lAddr=lAddr;
	  prg->_lLen=lLen;
	  prg->_f=lFlags;
	  pcb->_iMemRegion=i;
	}
	else {
	  /* IO space */
	  if (request_region(lAddr,lLen, pdm->_szDid)==0) {
	    printk(KERN_ERR PFX "io region #%d not available\n", pdm->_szDid, prd->_iBar);
	    goto lFail;
	  }

	  prg->_lAddr=lAddr;
	  prg->_lLen=lLen;
	  prg->_f=lFlags;

	  if(prd->_f & RG_AMCC5920_CONF) { 
	    outb(0x87, lAddr+0x60); 		// max(7)  waitstates for pass through configuration
	    outb(0x2c, lAddr+0x38+1); 	// set bit 13 in INTCSR to enable interrupts
	    printk(KERN_INFO PFX "AMCC5920 configured %lx\n", pdm->_szDid, lAddr);
	  } 
	  else if (prd->_f & RG_IOBASE) { 
	    pcb->_iIoRegion=i;
	  } 
	}
      }
      if (pcb->_iIoRegion == -1) goto lFail;

      printk(KERN_INFO PFX "adaptor type=%s manuf=%s arch=%s\n",pdm->_szDid, pad->_ad._szName, pad->_ad._szOrg, pad->_ad._szArch);

      /* request irq */
      pcb->_irq=pdv->irq;
      pcb->_fIrq=SA_SHIRQ;

      printk(KERN_INFO PFX "IRQ %d\n", pdm->_szDid, pcb->_irq);
      if (request_irq(pcb->_irq, busintr, pcb->_fIrq, pdm->_szDid, pcb) ) {
	printk(KERN_ERR PFX "IRQ %d not available\n", pdm->_szDid, pcb->_irq);
	goto lFail;
      }
    }
    busCardInitUmi(pdm, pcb, pArg);

    /*
     * register the card
     */
    pdm_setcard(pdm, iSlot, &pcb->_ca);
    add_timer(&pcb->_timer);
    return 0;

  lFail:
    busReleasePci(pcb);
    return (busCardDelete(pdm, pcb), -1);
  }

  void pdm_pci_remove(LumiMod *pdm, struct pci_dev *pdv) {
    char mpch[SZ_CARD_ID];
    int iSlot;
    if ((iSlot=pdm_findslot(pdm, szIdentPci(mpch, pdv)))!=-1) {
      CardBus*pcb;
      if ((pcb=(CardBus*)pdm_clrcard(pdm, iSlot))) {
	printk(KERN_INFO PFX "release pci card %s@%d\n", pdm->_szDid, mpch, iSlot);
	busReleasePci(pcb);
	busCardDelete(pdm, pcb);
      }
    }
  }





